﻿namespace KPGBase.Excercises
{
    internal class Excercise07 : BasicExcercise
    {
        /// <summary>  </summary>
        /// <summary> Path to lenna file </summary>
        public string kittyPath = "kitty2.jpg";
        /// <summary> Cell info </summary>
        int cellWidth = 0;
        int cellHeight = 0;
        private int NUMBER_OF_CELLS = 10;
        
        /// <summary> Original image </summary>
        private Bitmap original;
        /// <summary> Warped image </summary>
        private Bitmap changed;


        /// <summary> Control grid points </summary>
        Point[,] grid;           
        /// <summary> Shifts of the control grid points (how much did the points move from their original positions)
        /// - if the point is in its correct/original position the value is (0,0) </summary>
        Point[,] differences;    
        Point[,] gridImage;      // Obrazek ulozeny jako matice

        /// <summary> Attributes to detect user click  </summary>
        private const float MAX_DISTANCE = 5f; // Max distance to point to detect click
        Point startPoint; // coordinates where user clicked
        Point selectedPoint; // selected grid point the user clicked on
        bool leftClicked; // left mouse button clicked
        bool isSelected; // is a point selected? (did user click close enough to a grid point?)

        /// <summary>
        /// Constructor
        /// </summary>
        public Excercise07(Bitmap canvas) : base(canvas)
        {
            original = (Bitmap)canvas.Clone();
            changed = (Bitmap)canvas.Clone();
            LoadImage(kittyPath);
            SetDefaultGrid();
        }

        /// <summary>
        /// Set up default grid values
        /// </summary>
        private void SetDefaultGrid()
        {
            this.grid = InitGrid();
            this.differences = InitDifferences();
            this.gridImage = new Point[original.Width, original.Height];
            DrawGrid();
        }


        /// <summary>
        /// Load image from path and set it as local canvas + create graphics
        /// </summary>
        /// <param name="path"> Path to image </param>
        public void LoadImage(string path)
        {
            Bitmap? lenna = LoadOriginalImg(path);
            if (lenna != null)
            {
                canvas = lenna;
                original = (Bitmap)canvas.Clone();
                changed = (Bitmap)canvas.Clone();

                g = Graphics.FromImage(canvas);
            }
        }

        /// <summary>
        /// Load image from path
        /// </summary>
        /// <param name="path"> Path to image </param>
        /// <returns> Bitmap with loaded image </returns>
        private Bitmap? LoadOriginalImg(string path)
        {
            if (File.Exists(path))
                return (Bitmap)Image.FromFile(path);

            return null;
        }

        /// <summary>
        /// Init the control grid
        /// </summary>
        /// <returns></returns>
        private Point[,] InitGrid()
        {
            // Compute the points of the control grid and return them in a 2D array
            // - there should be NUMBER_OF_CELLS * NUMBER_OF_CELLS -> which means (NUMBER_OF_CELLS + 1)*(NUMBER_OF_CELLS + 1) control points
            // - first point is in the left top corner, last in the right bottom corner
            // - set cellWidth and cellHeight

            cellWidth = canvas.Width / NUMBER_OF_CELLS;
            cellHeight = canvas.Height / NUMBER_OF_CELLS;

            Point[,] matrix = new Point[NUMBER_OF_CELLS + 1, NUMBER_OF_CELLS + 1];
            for (int i = 0; i < NUMBER_OF_CELLS + 1; i++)
                for (int j = 0; j < NUMBER_OF_CELLS + 1; j++)
                    matrix[i, j] = new Point(i * cellWidth, j * cellHeight);
            
            return matrix;
        }

        /// <summary>
        /// Init differences = movement of control points
        /// </summary>
        /// <returns></returns>
        private Point[,] InitDifferences()
        {
            Point[,] matrix = new Point[NUMBER_OF_CELLS + 1, NUMBER_OF_CELLS + 1];
            
            for (int i = 0; i < NUMBER_OF_CELLS + 1; i++)
                for (int j = 0; j < NUMBER_OF_CELLS + 1; j++)
                    matrix[i, j] = new Point(0, 0);

            return matrix;
        }

        private void Redraw()
        {
            // Calculate offsets of image points
            CalcOffsets();   
            // Apply offsets -> create new image changed
            OffsetFilterSlow(changed, gridImage);
            // Draw image
            g.DrawImage(changed, new Rectangle(0, 0, original.Width, original.Height));
            // Draw grid
            DrawGrid(); 
        }

        /// <summary>
        /// Draw the control grid
        /// </summary>
        private void DrawGrid()
        {
            // Draw the control grid
            // Either draw the whole grid or without the border points
            // - aka without the top-most, bottom-most, left-most and right-most points
            // - those points cannot be moved anyways

            for (int i = 1; i < NUMBER_OF_CELLS - 1; i++)
            {
                for (int j = 1; j < NUMBER_OF_CELLS - 1; j++)
                {
                    g.FillEllipse(Brushes.White, new Rectangle((int)(grid[i, j].X - MAX_DISTANCE), (int)(grid[i, j].Y - MAX_DISTANCE), (int)(2 * MAX_DISTANCE), (int)(2 * MAX_DISTANCE)));
                    g.DrawLine(Pens.White, grid[i, j].X, grid[i, j].Y, grid[i, j + 1].X, grid[i, j + 1].Y);
                    g.DrawLine(Pens.White, grid[i, j].X, grid[i, j].Y, grid[i + 1, j].X, grid[i + 1, j].Y);
                }
            }
            g.FillEllipse(Brushes.White, new Rectangle((int)(grid[NUMBER_OF_CELLS-1, NUMBER_OF_CELLS-1].X - MAX_DISTANCE), (int)(grid[NUMBER_OF_CELLS - 1, NUMBER_OF_CELLS - 1].Y - MAX_DISTANCE), (int)(2 * MAX_DISTANCE), (int)(2 * MAX_DISTANCE)));


            for (int k = 1; k < NUMBER_OF_CELLS - 1; k++)
            {
                g.FillEllipse(Brushes.White, new Rectangle((int)(grid[NUMBER_OF_CELLS - 1, k].X - MAX_DISTANCE), (int)(grid[NUMBER_OF_CELLS - 1, k].Y - MAX_DISTANCE), (int)(2 * MAX_DISTANCE), (int)(2 * MAX_DISTANCE)));
                g.FillEllipse(Brushes.White, new Rectangle((int)(grid[k, NUMBER_OF_CELLS - 1].X - MAX_DISTANCE), (int)(grid[k, NUMBER_OF_CELLS - 1].Y - MAX_DISTANCE), (int)(2 * MAX_DISTANCE), (int)(2 * MAX_DISTANCE)));
                g.DrawLine(Pens.White, grid[NUMBER_OF_CELLS - 1, k].X, grid[NUMBER_OF_CELLS - 1, k].Y, grid[NUMBER_OF_CELLS - 1, k + 1].X, grid[NUMBER_OF_CELLS - 1, k + 1].Y);
                g.DrawLine(Pens.White, grid[k, NUMBER_OF_CELLS - 1].X, grid[k, NUMBER_OF_CELLS - 1].Y, grid[k + 1, NUMBER_OF_CELLS - 1].X, grid[k + 1, NUMBER_OF_CELLS - 1].Y);
            }

        }

        /// <summary>
        /// For each affected pixel in the image calculate
        /// </summary>
        private void CalcOffsets()
        {
            // Compute bilinear interpolation
            // - for each cell neighboring the selected point
            //      - for each pixel inside this cell
            //          - compute new offsets for each pixel which are then stored in 2D array gridImage
            //          - offsets of control points are stored in 2D array differences

            int x = selectedPoint.X;
            int y = selectedPoint.Y;
            double xfrac = (double)1.0 / cellWidth;
            double yfrac = (double)1.0 / cellHeight;

            for (int i = x - 1; i <= x; i++)
            {
                for (int j = y - 1; j <= y; j++)
                {
                    // for all points in the cell
                    for (int m = 0; m < cellWidth; m++)
                    {
                        for (int n = 0; n < cellHeight; n++)
                        {
                            // compute billinear interpolation
                            // here X and Y are the shifts

                            double xW = m * xfrac;
                            double yW = n * yfrac;

                            double diffX = differences[i, j].X * (1 - xW) * (1 - yW)
                                        + differences[i, j + 1].X * (1 - xW) * yW
                                        + differences[i + 1, j].X * xW * (1 - yW)
                                        + differences[i + 1, j + 1].X * xW * yW;

                            double diffY = differences[i, j].Y * (1 - xW) * (1 - yW)
                                         + differences[i, j + 1].Y * (1 - xW) * yW
                                         + differences[i + 1, j].Y * xW * (1 - yW)
                                         + differences[i + 1, j + 1].Y * xW * yW;

                            // image pixels saved with coordinates [width, height]
                            gridImage[i * cellWidth + m, j * cellHeight + n].X = (int)-diffX; 
                            gridImage[i * cellWidth + m, j * cellHeight + n].Y = (int)-diffY;  
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Compute warped image
        /// </summary>
        /// <param name="b"> Created warped image </param>
        /// <param name="offset"> Array with offsets of all pixels </param>
        public void OffsetFilterSlow(Bitmap b, Point[,] offset)
        {
            // Edit image b using the offset (in parameter offset)
            // - create a copy of original bitmap - this is the source for pixels' new color
            // - go through all cells in the neighborhood of the selected point -> 4 cells
            //      - index in control grid of the selected point is in attribute selectedPoint
            //      - width and height of a cell is in attributes cellWidth and cellHeight
            //      - for each pixel in every cell
            //          - get the offset - from parameter offset (this array contains the offsets for all pixels in bitmap)
            //          - get the new color of the pixel from original image - new color is located at pixel position+offset
            //          - set this color as the color of pixel at current pixel position in bitmap b

            int xOffset, yOffset;

            // Go through related pixels in all 4 neighboring cells
            for (int x = cellWidth * selectedPoint.X - cellWidth; x < cellWidth * selectedPoint.X + cellWidth; x++)
            {
                for (int y = cellHeight * selectedPoint.Y - cellHeight; y < cellHeight * selectedPoint.Y + cellHeight; y++)
                {
                    // get the offset
                    xOffset = offset[x, y].X;
                    yOffset = offset[x, y].Y;

                    if (y + yOffset >= 0 && y + yOffset < b.Height && x + xOffset >= 0 && x + xOffset < b.Width)
                    {
                        // get new color value
                        Color c = original.GetPixel((x + xOffset), (y + yOffset));
                        // set bitmap pixel to new color
                        b.SetPixel(x, y, c);
                    }
                }
            }
        }

        /// <summary>
        /// Mouse clicked
        /// </summary>
        /// <param name="newLoc"> Where was the mouse clicked - xy coordinates </param>
        internal void MouseClicked(Point newLoc)
        {
            startPoint = newLoc;
            leftClicked = true;

            // Go through control points
            // - if user clicked closer than MAX_DISTANCE away from one of them ->
            //      - startPoint is position of click
            //      - selectedPoint is point with i, j coordinates IN THE CONTROL GRID
            //            - x = row of control point matrix
            //            - y = column of control point matrix
            // 
            // !!! ATTENTION !!! we do not want to allow user to select the first or last row and column of the points !!!

            float x, y;
                for (int i = 1; i < NUMBER_OF_CELLS && !isSelected; i++)
                {
                    for (int j = 1; j < NUMBER_OF_CELLS; j++)
                    {
                        x = startPoint.X - grid[i, j].X;
                        y = startPoint.Y - grid[i, j].Y;

                        if (x * x + y * y < MAX_DISTANCE * MAX_DISTANCE)
                        {
                            isSelected = true;
                            selectedPoint = new Point(i, j);
                            startPoint = newLoc;
                            break;
                        }
                    }
                }
        }

        /// <summary>
        /// Mouse moved
        /// </summary>
        /// <param name="newLoc"> New position </param>
        /// <returns></returns>
        internal Bitmap MouseMove(Point newLoc)
        {
            // If left button clicked and a point is selected
            if (leftClicked && isSelected)
            {
                // move point in grid
                grid[selectedPoint.X, selectedPoint.Y] = newLoc;

                // set the differences
                differences[selectedPoint.X, selectedPoint.Y].X += newLoc.X - startPoint.X;
                differences[selectedPoint.X, selectedPoint.Y].Y += newLoc.Y - startPoint.Y;

                startPoint = newLoc;
                Redraw();
            }

            return canvas;
        }

        /// <summary>
        /// Mouse un-clicked
        /// </summary>
        internal void MouseUp()
        {
            leftClicked = false;
            isSelected = false;
        }
    }
}
